import tonkin.datvar
import tonkin.values
from typing import List, Dict, Union

def _splitPacket(rawPacket: bytearray, vars: List[tonkin.datvar.Datvar]) -> List[bytearray]:
	splitPacket = []
	for var in vars:
		splitPacket.append(bytearray(rawPacket[:var.getLength()]))
		rawPacket = rawPacket[var.getLength():]
	return splitPacket

def readRawPacket(rawPacket: bytearray, vars: List[tonkin.datvar.Datvar]) -> Dict[str, Union[int, float, bool]]:
	vals = _splitPacket(rawPacket, vars)
	packet = {}
	for index, var in enumerate(vars):
		varName = var.name
		value = tonkin.values.readRawValue(vals[index], var.dataType)
		packet[varName] = value
	return packet

def readPackets(packets: List[bytearray], vars: List[tonkin.datvar.Datvar]) -> List[Dict[str, Union[int, float, bool]]]:
	output = []
	for packet in packets:
		output.append(readRawPacket(packet, vars))
	return output



import tonkin.listops as lops
import tonkin.datvar as datvar
from typing import List, Union
from countach import measurement as mea
from countach import characteristic as cha

# Split the variables section into individual definitions
def splitVarSectionToDefs(varSection: bytearray) -> List[List[int]]:
	return lops.splitListBySub(list(varSection), [0x43, 0x43, 0x3E])

# Extract a variable name from a definition
def extractVariableName(varDef: List[int]) -> str:
	outputBytes = []
	# Remove data before variable name
	varDef = varDef[66:]

	# Extract name
	for i in varDef:
		if i == 0x00:
			break
		outputBytes.append(chr(i))

	return "".join(outputBytes)

# Get a list of variable names from a list of definitions (ignores the last one)
def extractVariableNames(varDefs: List[List[int]]) -> List[str]:
	# Don't get the "time" variable
	varDefs.pop(-1)
	output = []
	for var in varDefs:
		output.append(extractVariableName(var))
	return output

# Given a variable name and the A2L data, return a datvar object representing the variable
def getDatVarFromA2L(variableName: str, a2lData: List[Union[mea.Measurement, cha.Characteristic]]) -> datvar.Datvar:
	for i in a2lData:
		rightCategory = type(i) == mea.Measurement
		rightName = i.name == variableName
		if rightCategory and rightName:
			return datvar.Datvar(variableName, i.dataType)
	raise ValueError("Variable " + variableName + " not found in A2L data")

def getDatVarsFromA2L(variableNames: List[str], a2lData: List[Union[mea.Measurement, cha.Characteristic]]) -> List[datvar.Datvar]:
	output = []
	for variableName in variableNames:
		output.append(getDatVarFromA2L(variableName, a2lData))
	return output

def getVariableListFromA2L(varSection: bytearray, a2lData: List[Union[mea.Measurement, cha.Characteristic]]) -> List[datvar.Datvar]:
	defs = splitVarSectionToDefs(varSection)
	variableNames = extractVariableNames(defs)
	variableList = getDatVarsFromA2L(variableNames, a2lData)
	variableList.append(datvar.Datvar("time", "single"))
	return variableList



import tonkin.listops
import tonkin.datvar
from typing import List

def _calculatePacketLength(vars: List[tonkin.datvar.Datvar]) -> int:
	length = 0
	for var in vars:
		length += var.getLength()
	return length

def splitStreamToPackets(stream: bytearray, vars: List[tonkin.datvar.Datvar]):
	packetLength = _calculatePacketLength(vars)
	return tonkin.listops.splitListIntoLengths(list(stream), packetLength)



import countach.parse
from tonkin import datvar, varsec, datfile

# Set up the test datvars in the right order with the extra "time" one at the end
def getTestDatVars():
	tv = countach.parse.parseFile("tests/testdata/test_0.a2l")
	datvars = []
	singleval = tv[8]
	int32val = tv[9]
	uint16val = tv[10]
	int16val = tv[11]
	fixedval = tv[12]
	uint8val = tv[13]
	int8val = tv[14]
	boolval = tv[15]

	testVariables = [uint8val, uint16val, singleval, int8val, int32val, int16val, fixedval, boolval]

	for i in testVariables:
		datvars.append(datvar.Datvar(i.name, i.dataType))
	datvars.append(datvar.Datvar("time", "single"))
	return datvars

testVariables = getTestDatVars()

def test_getVariableListFromA2L():
	global testVariables
	a2ldata = countach.parse.parseFile("tests/testdata/test_0.a2l")
	_, variableSection, _ = datfile.getSectionsFromFile("tests/testdata/multitype.dat")
	realVars = varsec.getVariableListFromA2L(variableSection, a2ldata)
	assert realVars == testVariables



from typing import Tuple, Union

def isSigned(typeStr: str) -> bool:
	if typeStr[0] == "u":
		signed = False
	elif typeStr[0] == "s":
		signed = True
	else:
		raise ValueError("Sign indication not found in type string")
	return signed

def getWordLength(typeStr: str) -> int:
	typeStr = typeStr[4:]
	splitTypeStr = typeStr.split("_")
	return int(splitTypeStr[0])

def getTotalSlope(typeStr: str) -> int:
	totalSlope = 1
	if "S" in typeStr:
		totalSlopeIndex = typeStr.index("S") + 1
		totalSlope = int(typeStr[totalSlopeIndex])
	elif "E" in typeStr:
		totalSlopeIndex = typeStr.index("E") + 1
		totalSlope = 2 ** int(typeStr[totalSlopeIndex])
	return totalSlope

def getBias(typeStr: str) -> int:
	bias = 0
	if "B" in typeStr:
		biasIndex = typeStr.index("B") + 1
		bias = int(typeStr[biasIndex])
	return bias

def getPropertiesFromFixedTypeStr(typeStr: str) -> Tuple[bool, int, int, int]:
	signed = isSigned(typeStr)
	wordLength = getWordLength(typeStr)
	totalSlope = getTotalSlope(typeStr)
	bias = getBias(typeStr)
	return signed, wordLength, totalSlope, bias

# Round n up to the nearest multiple
def roundUp(n: Union[int, float, bool], multiple: Union[int, float, bool]) -> Union[int, float, bool]:
	if multiple == 0:
		return n

	remainder = n % multiple

	if remainder == 0:
		return n

	return n + multiple - remainder

def getByteLength(bitLength: int) -> int:
	totalBits = roundUp(bitLength, 8)
	return int(totalBits / 8)

def getByteLengthFromFixedTypeStr(typeStr: str) -> int:
	_, wordLength, _, _ = getPropertiesFromFixedTypeStr(typeStr)
	return getByteLength(wordLength)

# Remove the extra bits from the start of a fixdt value
def chopExtraBits(value: int, wordLength: int, byteLength: int) -> int:
	# The case where the wordLength is a multiple of 8
	if wordLength / 8 == byteLength:
		return value
	
	bitsToChop = (byteLength * 8) - wordLength
	binary = bin(value)
	# Chop an extra 2 chars for "0b"
	chopped = binary[bitsToChop + 2:]
	return int(chopped)



from dataclasses import dataclass
import tonkin.fixed

# A variable as it appears in the header section
@dataclass
class Datvar:
	name: str
	dataType: str

	def getLength(self) -> int:
		if "fix" in self.dataType:
			return tonkin.fixed.getByteLengthFromFixedTypeStr(self.dataType)
		elif "int8" in self.dataType:
			return 1
		elif "int16" in self.dataType:
			return 2
		elif "int32" in self.dataType:
			return 4
		elif self.dataType == "single":
			return 4
		elif self.dataType == "boolean":
			return 1
		elif self.dataType == "time":
			return 4
		else:
			raise ValueError(f"{self.dataType} is not a valid type string")



import struct
from typing import Union

def convertint8(int8B: bytearray) -> int:
	return int.from_bytes(int8B, "little", signed=True)

def convertint16(int16B: bytearray) -> int:
	return int.from_bytes(int16B, "little", signed=True)

def convertint32(int32B: bytearray) -> int:
	return int.from_bytes(int32B, "little", signed=True)

def convertuint8(uint8B: bytearray) -> int:
	return int.from_bytes(uint8B, "little", signed=False)

def convertuint16(uint16B: bytearray) -> int:
	return int.from_bytes(uint16B, "little", signed=False)

def convertSingle(singleB: bytearray) -> float:
	return float(struct.unpack("f", singleB)[0])

def convertBoolean(boolB: bytearray) -> bool:
	if boolB == b'\x01':
		return True
	else:
		return False

# Given a series of bytes representing a value within a packet
# and a variable dict, return the value of the 
def readRawValue(rawValue: bytearray, varType: str) -> Union[int, float, bool]:
	if varType == "int8":
		return convertint8(rawValue)
	elif varType == "int16":
		return convertint16(rawValue)
	elif varType == "int32":
		return convertint32(rawValue)
	elif varType == "uint8":
		return convertuint8(rawValue)
	elif varType == "uint16":
		return convertuint16(rawValue)
	elif varType == "single":
		return convertSingle(rawValue)
	elif varType == "boolean":
		return convertBoolean(rawValue)
	elif "fix" in varType:
		raise NotImplementedError("Fixed-point decoding not supported yet")
	else:
		raise ValueError("No valid type found in varType")



import countach.parse
import tonkin.fileops as fops
import tonkin.verification
import tonkin.datfile
import tonkin.varsec
import tonkin.datastream
import tonkin.packets
import csv
from typing import List, Dict, Tuple, Union

def readDatFileWithA2L(datFile: str, a2lFile: str) -> List[Dict[str, Union[int, float, bool]]]:
	# Extract the data from the file
	a2lData = countach.parse.parseFile(a2lFile)
	fileBytes = fops.importFileAsBytes(datFile)
	headerSection, varSection, streamSection = tonkin.datfile.splitFileBytes(fileBytes)

	# Verify the file
	headerStatus = tonkin.verification.checkHeader(headerSection)
	if headerStatus == False:
		raise ValueError("File is not in valid EcoCAL .dat format")

	# Extract the data
	vars = tonkin.varsec.getVariableListFromA2L(varSection, a2lData)
	rawPackets = tonkin.datastream.splitStreamToPackets(streamSection, vars)
	return tonkin.packets.readPackets(rawPackets, vars)

# Write the data to a CSV file
def datToCSVWithA2L(datFile: str, a2lFile: str, outputFile: str):
	data = readDatFileWithA2L(datFile, a2lFile)
	headers = data[0].keys()
	with open(outputFile, "w") as csvFile:
		writer = csv.DictWriter(csvFile, fieldnames=headers)
		writer.writeheader()
		for row in data:
			writer.writerow(row)

def datTo2DWithA2L(datFile: str, a2lFile: str) -> List[List[Union[int, float, bool]]]:
	data = readDatFileWithA2L(datFile, a2lFile)
	headers = list(data[0].keys())
	output = []
	output.append(headers)
	for i in data:
		row = [i.get(key, "") for key in headers]
		output.append(row)
	return output



import tonkin.listops as lops
import tonkin.fileops as fops
from typing import Tuple

def extractHeaderSection(fileBytes: bytearray) -> bytearray:
	return fileBytes[0:228]

def extractVariablesSection(fileBytes: bytearray) -> bytearray:
	withoutHeader = fileBytes[228:]
	end = lops.getSubListIndex(list(withoutHeader), [0x43, 0x47, 0x1A])
	return withoutHeader[:end]

def extractDataSection(fileBytes: bytearray) -> bytearray:
	headerLength = len(extractHeaderSection(fileBytes))
	variableLength = len(extractVariablesSection(fileBytes))
	upperLength = headerLength + variableLength
	output = fileBytes[upperLength:]
	# The sequence of bytes that marks the start of the data stream
	startMarker = [0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]
	# The index of the first byte of the data stream
	start = lops.getSubListIndex(list(output), list(startMarker)) + 1
	return output[start:]

def splitFileBytes(fileBytes: bytearray) -> Tuple[bytearray, bytearray, bytearray]:
	headerSection = extractHeaderSection(fileBytes)
	variableSection = extractVariablesSection(fileBytes)
	dataSection = extractDataSection(fileBytes)
	return headerSection, variableSection, dataSection

def getSectionsFromFile(fileName: str) -> Tuple[bytearray, bytearray, bytearray]:
	fileBytes = fops.importFileAsBytes(fileName)
	return splitFileBytes(fileBytes)



def verifyASCII(source: bytearray, result: str, start: int, stop: int) -> bool:
	return source[start:stop].decode("ASCII") == result

# Check if the file uses the MDF format
def checkFormat(header: bytearray) -> bool:
	return verifyASCII(header, "MDF", 0, 3)

# Check if the file uses MDF 3.00
def checkVersion(header: bytearray) -> bool:
	return verifyASCII(header, "3.00", 8, 12)

# Check if the vendor is Ecotrons
def checkVendor(header: bytearray) -> bool:
	return verifyASCII(header, "Ecotrons", 16, 24)

# Check if the file was created by EcoCAL
def checkSoftware(header: bytearray) -> bool:
	return verifyASCII(header, "EcoCAL", 100, 106)

# Check that the second vendor is Ecotrons
def checkSecondVendor(header: bytearray) -> bool:
	return verifyASCII(header, "Ecotrons", 132, 140)

def checkHeader(header: bytearray) -> bool:
	if not checkFormat(header):
		return False
	elif not checkVersion(header):
		return False
	elif not checkVendor(header):
		return False
	elif not checkSoftware(header):
		return False
	elif not checkSecondVendor(header):
		return False
	return True



from typing import List, Any, Iterator

def shiftListLeft(input: List[Any], shift: int) -> List[Any]:
	# Bring large numbers down to smaller equivalents
	# and handle negatives for shifts right
	shift = shift % len(input)

	for i in range(0, shift):
		input.append(input.pop(0))
	return input

# Given a source list and a target list, returns the index of where the target
# list appears within the source list (returns the index of the final value)
def getSubListIndex(source: List[Any], target: List[Any]) -> int:
	# TODO: Ensure that the buffer items don't appear in split
	buffer = [None] * len(target)
	
	for j, i in enumerate(source):
		buffer = shiftListLeft(buffer, 1)
		buffer[-1] = i

		if buffer == target:
			return j
	
	raise ValueError("Sub-list not found")

def getSubListIndices(source: List[Any], target: List[Any]) -> List[int]:
	output = []

	# TODO: Ensure that the buffer items don't appear in split
	buffer = [None] * len(target)
	
	for j, i in enumerate(source):
		buffer = shiftListLeft(buffer, 1)
		buffer[-1] = i

		if buffer == target:
			output.append(j)
	
	return output

# TODO: Test for different first elements of source
def splitListBySub(source: List[Any], target: List[Any]) -> List[List[Any]]:
	output = []
	indices = getSubListIndices(source, target)
	
	# The case where the target does not appear
	if indices == []:
		return [source]

	for j, i in enumerate(indices):
		startOfBlock = i - len(target) + 1
		
		if j < len(indices) - 1:
			endOfBlock = indices[j+1] - len(target) + 1
		else:
			endOfBlock = len(source)

		output.append(source[startOfBlock:endOfBlock])

	return output

def splitListIntoLengthsGen(input: List[Any], lengths: int) -> Iterator[List[Any]]:
	for i in range(0, len(input), lengths):
		yield input[i:i + lengths]

def splitListIntoLengths(input: List[Any], lengths: int) -> List[Any]:
	return list(splitListIntoLengthsGen(input, lengths))
